#!/usr/bin/env python
# -*- Encoding: latin-1 -*-
"""
A class for handling franctions
written 2004 by Marek Kubica

Uses code from Mike Hostetler's fraction.py, but is more powerful
and is _not_ an extension of fraction.py, but uses some concepts of these.
Licensed under LGPL

Differences between mh and mk fraction implementation:
- mk does not work (will be fixed)
- mk can work with whole numbers
- mk version is more documented
- mk has unitttests, and is in big parts test-first developed
- mh seems to be more straightforward
- the performance was not yet tested (but has to be done)

TODO:
- make more unit tests (priority: as needed)
- look throught the unittest to find dublicates (priority: low)
- let fraction work with simple operations (priority: high)
- keep Fraction immutable
- tune performance (lesser calls to __mixed(), or make it faster, maybe by a state variable)
  (priority: low)

Changelog:
0.0.10 - immutability reached: coded 'lower' handlers, made __input() and __getValue() (finaly)
	private, coded getUnrealNumber()
0.0.9 - implemented reducing, first steps towards immutability: coded 'whole' and 'upper' handlers,
	observed slowdown due to __mixed() calls
0.0.8 - made some more tests, made better and more comments
0.0.7 - made getValue() work better, improved input(), implemented __unreal and __mixed
0.0.6 - namespace mangling (means semi-private internal functions)
0.0.5 - __repr__, __float__, __int__ hooks, more unittests
0.0.4 - made all unittests pass"""

__version__ = '0.0.10'
__author__ = 'Marek Kubica'

class Fraction(object):
    """This class is intended to handle fractions, to operate with them
    yust like with normal numbers"""
    
    #__slots__ = ['whole', 'upper', 'lower']
    
    def __init__(self, whole=0, upper=0, lower=1):
        """The constructor of the Fraction
        this method is overloaded (well, kind of):
        it can be called by
        Fraction(whole=1, upper=1, lower=2)
        or by
        Fraction("a b/c")
        (well Fraction(whole="a b/c") would also work but is illogical)
        """
        if type(whole) == type(str()) or type(whole) == type(unicode()):
            # alternate constructor - 'a b/c' - string constructor
            self.__input(whole)
        else:
            # traditional constructor - eg. whole=1, upper=1, lower=2
            self.__setWhole(whole)
            self.__setUpper(upper)
            self.__setLower(lower)

    def __repr__(self):
	"""The 'print' representation of the fraction.
	This is also the hook for the repr() & str() builtin.
	Uses self.getMixedNumber() internally."""
	return self.getMixedNumber()

    def __float__(self):
	"""Returns the fraction as float.
	This is the hook for the float() builtin."""
        # well, this *could* be made shorter, but now it's logical and easy
        
	# swich to mixed mode
	self.__mixed()
        # calculate the fraction part
        return_float = float(self.__getUpper()) / float(self.__getLower())
        # add the whole part
        return_float = return_float + self.__getWhole()
        # return the result
        return return_float

    def __int__(self):
	"""Returns just the whole part of the fraction.
	This is the hook for the int() builtin."""
	# swich to mixed mode
	self.__mixed()
	return self.__getWhole()

    def __add__(self, fract):
	raise NotImplementedError('First make this immutable')

    def __mixed(self):
	"""This function converts the mixed number fraction into a unreal fraction"""
	# calling __unreal makes it possible to work also with fraction with whole numbers
	self.__unreal()
	# so we don't need to care about the 'whole' values
	
	# getting the new values: we don't need whole, since 
	#+calling __unreal makes it 0 (and increases upper)
	u, l = self.__getUpper(), self.__getLower()
	
	# the new whole is the int value of upper / lower (floordivision?)
	w_new = u / l
	
	# How much do we need to substract from upper? 
	#+The new whole multiplied with lower!
	u_less = w_new * l
	
	# Und substract it finally
	u_new = u - u_less
	
	# set the new whole and upper values
	self.__setWhole(w_new)
	self.__setUpper(u_new)
	# (lower stays the same, this should be so!)
	
	
    def __unreal(self):
	"""This function converts unreal fractions into mixed number fractions"""
	w, u, l = self.__getWhole(), self.__getUpper(), self.__getLower()
	
	# we just need to run it if whole is not 0
	#+else is 1) is already an unreal fraction 2) is a normal fraction < 1
	if w != 0:
	    # set upper to lower multiplied with whole and added to upper
	    self.__setUpper(l * w + u)
	    # and set whole to zero (since it is now in upper)
	    self.__setWhole(0)
    
    def __input(self, fraction):
        """Input strings in the form 'a b/c'"""
	
	# first split the given string on a space
        fraction = fraction.split(' ')
	
	# now look how many chunks we got
	if len(fraction) != 2:
	    # we got a smaller fraction: just a whole or just a fraction
	    tempval = fraction[0]
	    # look if there is a slash
	    if tempval.count('/') == 0:
		# no slash, we got just a whole
		whole = int(tempval)
		self.__setWhole(whole)
		
		# set the other values (effectively to zero)
		self.__setUpper(0)
		self.__setLower(1)
		
	    else:
		# we got a slash so we have a fraction part
		#+so split it by the slash
		fract_part = tempval.split('/')
		
		# and set the values
		upper = int(fract_part[0])
		lower = int(fract_part[1])
		self.__setUpper(upper)
		self.__setLower(lower)
		
		# set Whole to zero
		self.__setWhole(0)
	else:
	    # we got a full size fraction: with whole and fraction part
	    # set the whole value
    	    whole = int(fraction[0])
    	    self.__setWhole(whole)
	    
	    # now split the rest by the slash...
    	    fract_part = fraction[1].split('/')
	    # ...and set the rest of the values
    	    upper = int(fract_part[0])
    	    lower = int(fract_part[1])
    	    self.__setUpper(upper)
    	    self.__setLower(lower)

    def __getPrettyprint(self):
	"""The mk prettyprint mechanism.
	This could surely be improved so that the upper and lower values
	stand in the middle of the line, not just at the beginning"""
	self.__mixed()
	w, u, l = self.__getWhole(), self.__getUpper(), self.__getLower()
	
	if float(u) / float(l) != 0:
	    # get the lenghts of the numbers
	    lw, lu, ll = len(str(w)), len(str(u)), len(str(l))
	    
	    # create the line
	    if lu > ll:
		line = lu * '-'
	    else:
		line = ll * '-'
	    
	    # create the prettyprinted lines
	    pline1 = (lw + 1) * ' ' + str(u)
	    pline2 = str(w) + ' ' + line
	    pline3 = (lw + 1) * ' ' + str(l)
	    
	    # make them to one prettyprinted multiline string
	    pprint = pline1 + '\n' + pline2 + '\n' + pline3
	    # return the string
	    return pprint
	else:
	    # we have no fraction part, so just return the whole part
	    return str(w)

    
    def prettyprint(self):
	"""A small prettyprint wrapper"""
	print self.__getPrettyprint()
        
    def __getWhole(self):
        """Returns the whole part of the fraction"""
        return self.whole
    
    def __getUpper(self):
	"""Returns the demominator of the fraction"""
        return self.upper
    
    def __getLower(self):
	"""Returns the numerator of the fraction"""
        return self.lower
    
    def __setWhole(self, number):
        """Set the 'whole' value.
	This function does checking and if the value is bad raises ValueError."""
        if type(number) == type(int()) or type(number) == type(float()):
            self.whole = number
        else:
            raise ValeError('A number is expected')
            
    def __setUpper(self, number):
        try:
            int(number)
        except (ValueError, TypeError):
            raise ValueError('A number is expected')
        else:
            self.upper = number
    
    def reduce(self):
	"""Does actually run reduceing.
	Should this take care of the reduce switch? I think not!
	However this is not yet implemented, maybe I can use code from
	the mh fraction implementation"""
        #raise NotImplementedError

	divisor = self.__gcd(self.__getUpper(), self.__getLower())
	if divisor > 1:
	    upper = self.__getUpper() / divisor
	    lower = self.__getLower() / divisor
	    self.__setUpper(upper)
	    self.__setLower(lower)
    
    def __setLower(self, number):
        """Sets the value of Lower.
	This function checks the value before setting any bad and raises a nice ValueError."""
        # should we also accept string numbers? No! (isn't it?)
	#+but we still do: we convert them with int()!!
        
        # check if input _is_ a number
        try:
            int(number)
        except ValueError, TypeError:
            raise ValueError('A number is expected')
        
        # okay, this is a number, check if the number is zero
        if number == 0:
            raise ValueError('A zero value is not allowed')
        else:
            self.lower = number
    
    def __getValue(self):
        """Returns the simple string representation of the fraction.
	If the fraction part is zero it returns just the whole part.
	This function is deprecated to use externally. It is used
	by other get* Functions but later it may disappear, because it
	shows the *real* internal state of the fraction (and it is
	not predictable how it will look like)"""
        w, u, l = self.__getWhole(), self.__getUpper(), self.__getLower()
	
	if float(u) / float(l) != 0.0:
	    if w != 0:
    		string_output = str(w) + ' ' + str(u) + '/' + str(l)
    		return string_output
	    else:
		string_output = str(u) + '/' + str(l)
		return string_output
	else:
	    string_output = str(w)
	    return string_output

    def getMixedNumber(self):
	"""Returns the fraction as a mixed number.
	The implementation seems a bit like a dirty hack."""
	self.__mixed()
	return self.__getValue()
	
    def getUnrealNumber(self):
	"""Returns the fraction as an unreal fraction.
	This is the same type of hack like in getMixedNumber()"""
	self.__unreal()
	return self.__getValue()
    
    def __gcd(self, a, b):
	"""Helper method for reducing.
	Don't ask me what this *really* makes, 
	the code is taken from mh fraction"""
	if b==0:
	    return a
	else:
	    return self.__gcd(b, a%b)

def extractWhole(fraction):
    """This function is for extracting the value of whole in a fraction.
    This is useful only in newer fraction versions, since in these versions
    access to plain variables is deprecated."""
    if fraction.count(' ') == 0:
	# we have no whole (means whole is zero)
	return 0
    else:
	# split the fraction and return the first part
	fraction = fraction.split(' ')
	return int(fraction[0])

def extractUpper(fraction):
    """This ........"""
    if fraction.count(' ') == 1:
	# we got a two part fraction
	fraction = fraction.split(' ')
	fract_part = fraction[1].split('/')
	return int(fract_part[0])
    else:
	# a one part fraction
	if fraction.count('/') == 1:
	    fract_part = fraction.split('/')
	    return int(fract_part[0])
	else:
	    # just a whole value - no (real) upper value
	    return 0

def extractLower(fraction):
    """This ........
    Is almost exactly the same code as extractUpper()"""
    if fraction.count(' ') == 1:
	# we got a two part fraction
	fraction = fraction.split(' ')
	fract_part = fraction[1].split('/')
	return int(fract_part[1])
    else:
	# a one part fraction
	if fraction.count('/') == 1:
	    fract_part = fraction.split('/')
	    return int(fract_part[1])
	else:
	    # just a whole value - dummy lower value
	    # we cannot set 0 here, because lower *never* is zero
	    return 1